/* AXPbox Alpha Emulator
 * Copyright (C) 2020 Tomáš Glozar
 * Website: https://github.com/lenticularis39/axpbox
 *
 * Forked from: ES40 emulator
 * Copyright (C) 2007-2008 by the ES40 Emulator Project
 * Copyright (C) 2007 by Camiel Vanderhoeven
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 *
 * Although this is not required, the author would appreciate being notified of,
 * and receiving any modifications you may make to the source code that might
 * serve the general public.
 *
 * Parts of this file based upon the Poco C++ Libraries, which is Copyright (C)
 * 2004-2006, Applied Informatics Software Engineering GmbH. and Contributors.
 */

//
// Semaphore.h
//
// $Id: Semaphore.h,v 1.1 2008/05/31 15:47:28 iamcamiel Exp $
//
// Library: Foundation
// Package: Threading
// Module:  Semaphore
//
// Definition of the Semaphore class.
//
// Copyright (c) 2004-2006, Applied Informatics Software Engineering GmbH.
// and Contributors.
//
// Permission is hereby granted, free of charge, to any person or organization
// obtaining a copy of the software and accompanying documentation covered by
// this license (the "Software") to use, reproduce, display, distribute,
// execute, and transmit the Software, and to prepare derivative works of the
// Software, and to permit third-parties to whom the Software is furnished to
// do so, all subject to the following:
//
// The copyright notices in the Software and this entire statement, including
// the above license grant, this restriction and the following disclaimer,
// must be included in all copies of the Software, in whole or in part, and
// all derivative works of the Software, unless such copies or derivative
// works are solely in the form of machine-executable object code generated by
// a source language processor.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
// FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//

#ifndef Foundation_Semaphore_INCLUDED
#define Foundation_Semaphore_INCLUDED

#include "Exception.hpp"
#include "Foundation.hpp"

#if defined(POCO_OS_FAMILY_WINDOWS)
#include "Semaphore_WIN32.hpp"
#else
#include "Semaphore_POSIX.hpp"
#endif

class CSemaphore : private CSemaphoreImpl
/// A Semaphore is a synchronization object with the following
/// characteristics:
/// A semaphore has a value that is constrained to be a non-negative
/// integer and two atomic operations. The allowable operations are V
/// (here called set()) and P (here called wait()). A V (set()) operation
/// increases the value of the semaphore by one.
/// A P (wait()) operation decreases the value of the semaphore by one,
/// provided that can be done without violating the constraint that the
/// value be non-negative. A P (wait()) operation that is initiated when
/// the value of the semaphore is 0 suspends the calling thread.
/// The calling thread may continue when the value becomes positive again.
{
public:
  CSemaphore(int n);
  CSemaphore(int n, int max);
  /// Creates the semaphore. The current value
  /// of the semaphore is given in n. The
  /// maximum value of the semaphore is given
  /// in max.
  /// If only n is given, it must be greater than
  /// zero.
  /// If both n and max are given, max must be
  /// greater than zero, n must be greater than
  /// or equal to zero and less than or equal
  /// to max.

  ~CSemaphore();
  /// Destroys the semaphore.

  void set();
  /// Increments the semaphore's value by one and
  /// thus signals the semaphore. Another thread
  /// waiting for the semaphore will be able
  /// to continue.

  void wait();
  /// Waits for the semaphore to become signalled.
  /// To become signalled, a semaphore's value must
  /// be greater than zero.
  /// Decrements the semaphore's value by one.

  void wait(long milliseconds);
  /// Waits for the semaphore to become signalled.
  /// To become signalled, a semaphore's value must
  /// be greater than zero.
  /// Throws a TimeoutException if the semaphore
  /// does not become signalled within the specified
  /// time interval.
  /// Decrements the semaphore's value by one
  /// if successful.

  bool tryWait(long milliseconds);
  /// Waits for the semaphore to become signalled.
  /// To become signalled, a semaphore's value must
  /// be greater than zero.
  /// Returns true if the semaphore
  /// became signalled within the specified
  /// time interval, false otherwise.
  /// Decrements the semaphore's value by one
  /// if successful.

private:
  CSemaphore();
  CSemaphore(const CSemaphore &);
  CSemaphore &operator=(const CSemaphore &);
};

//
// inlines
//
inline void CSemaphore::set() { setImpl(); }

inline void CSemaphore::wait() { waitImpl(); }

inline void CSemaphore::wait(long milliseconds) {
  if (!waitImpl(milliseconds))
    throw CTimeoutException();
}

inline bool CSemaphore::tryWait(long milliseconds) {
  return waitImpl(milliseconds);
}

#endif // Foundation_Semaphore_INCLUDED
